﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Globalization;
using System.Security.Policy;
using System.Reflection;
using System.Diagnostics.CodeAnalysis;


namespace AppDomainLoadUnloadEx2
{
    /// <summary>
    /// Loads an assembly into a new AppDomain and obtains all the namespaces in the loaded Assembly, which are returned as a 
    /// List. The new AppDomain is then Unloaded.
    /// 
    /// This class creates a new instance of a <c>AssemblyLoader</c> class which does the actual ReflectionOnly loading of the Assembly into
    /// the new AppDomain.
    /// </summary>
    public class SeparateAppDomainAssemblyLoader
    {
        public AppDomain childDomain;

        #region Public Methods
        /// <summary>
        /// Loads an assembly into a new AppDomain and obtains all the namespaces in the loaded Assembly, which are returned as a 
        /// List. The new AppDomain is then Unloaded 
        /// </summary>
        /// <param name="assemblyLocation">The Assembly file 
        /// location</param>
        /// <returns>A list of found namespaces</returns>
        public List<String> LoadAssemblies(List<FileInfo> assemblyLocations, string topLevelNamespace)
        {
            List<String> namespaces = new List<String>();
 
            childDomain = BuildChildDomain(AppDomain.CurrentDomain);
 
            try
            {
                Type loaderType = typeof(AssemblyLoader);

                if (loaderType.Assembly != null)
                {
                    var loader = 
                        (AssemblyLoader)childDomain.
                            CreateInstanceFrom(
                            loaderType.Assembly.Location, 
                            loaderType.FullName).Unwrap();  // CreateInstanceFrom returns a System.Runtime.Remoting.ObjectHandle.  According to MSDN, this is 
                                                            // an object that is a wrapper for the new instance, or null if the type is not found.  The return
                                                            // value needs to be unwrapped to access the real object.
                                                            // 
                                                            // http://msdn.microsoft.com/en-us/library/2xkww633(v=vs.110).aspx

                    namespaces = loader.LoadAssemblies(assemblyLocations, topLevelNamespace);
                }

                return namespaces;
            }
            finally
            {
                AppDomain.Unload(childDomain);
            }
        }
        #endregion
 
        #region Private Methods
        /// <summary>
        /// Creates a new AppDomain based on the parent AppDomains' Evidence and AppDomainSetup
        /// </summary>
        /// <param name="parentDomain">The parent AppDomain</param>
        /// <returns>A newly created AppDomain</returns>
        private AppDomain BuildChildDomain(AppDomain parentDomain)
        {
            Evidence evidence = new Evidence(parentDomain.Evidence);
            AppDomainSetup setup = parentDomain.SetupInformation;
            return AppDomain.CreateDomain("MainProc.ChildAppDomain", evidence, setup);
        }
        #endregion
 
 
        /// <summary>
        /// Remotable AssemblyLoader, this class inherits from <c>MarshalByRefObject</c> 
        /// to allow the CLR to marshall this object by reference across AppDomain boundaries
        /// </summary>
        class AssemblyLoader : MarshalByRefObject
        {
            #region Private/Internal Methods
            /// <summary>
            /// ReflectionOnlyLoad of single Assembly based on 
            /// the assemblyPath parameter
            /// </summary>
            /// <param name="assemblyLocations"></param>
            /// <returns></returns>
            [SuppressMessage("Microsoft.Performance", "CA1822:MarkMembersAsStatic")]
            internal List<String> LoadAssemblies(List<FileInfo> assemblyLocations, string topLevelNamespace)
            {
                var namespaces = new List<String>();

                try
                {
                    foreach (FileInfo assemblyLocation in assemblyLocations)
                    {
                        System.Diagnostics.Debug.WriteLine(assemblyLocation.FullName);
                        Assembly.LoadFrom(assemblyLocation.FullName);
                    }
 
                    foreach (Assembly hostedAssembly in AppDomain.CurrentDomain.GetAssemblies())
                    {
                        foreach (Type type in hostedAssembly.GetTypes())
                        {
                            if ((type.Namespace != null) && (type.Namespace.StartsWith(topLevelNamespace)))
                            {
                                namespaces.Add(hostedAssembly.GetName().Name + ": " + type.FullName);
                            }
                        }                   
                    }

                    return namespaces;
                }
                catch (FileNotFoundException)
                {
                    // Just return the namespaces collected to this point.
                    //
                    return namespaces;
                }
            }
            #endregion
        }
    }
}

